/*
 * Implementation of an interactive theorem prover based on the use of
 * tactics and tacticals. This is the main program in this directory.
 */

module inter.

accum_sig  hyll, atom, hyllprf, goaltypes, hyllgoal, hylltac, goalred,
           tacticals, hylltac, examples, listmanip.

accumulate hylltac, tacticals, goalred, examples, listmanip.

/*
Next steps:
1) implement the remaining inference rules as tactics
2) implement the remaining clauses of the program that converts a proof
   to a string used for input to Coq.  (See write_proof below.)
*/

/* ********************************************************************
 * Top level predicates
 * p_inter: used to start a proof from the beginning, the user must
 *          store the input sequence of tactics in a file called
 *          "Name".hyll in order to continue the proof later using
 *          the p predicate.
 * p: used to continue a proof started by p_inter using the input from
 *    "Name".hyll, and then when completed, translate the complete
 *    proof to Coq script, which is printed to output file "Name".v. 
 ******************************************************************** */
type	p_inter		sname -> (int -> prf) -> (int -> goal) -> o.
type    p		sname -> (int -> goal) -> (int -> prf) -> o.

p_inter Name P OutGoal :- asequent Name Seq,
  pi w\ inter ((P w) # (Seq w)) (OutGoal w).

p Name OutGoal P :- asequent Name Seq, term_to_string Name Str,
  pi w\ (continue_proof Str ((P w) # (Seq w)) (OutGoal w),
         (Seq w) = (seq (Sig w) (G w) (D w) (J w)),
         write_proof (P w) (J w) Str).

/* ********************************************************************
 * Top level predicates for specific examples
 ******************************************************************** */

type    p1		(int -> goal) -> (int -> prf) -> o.
type    p1a		(int -> goal) -> (int -> prf) -> o.
type    p1b		(int -> goal) -> (int -> prf) -> o.
type	p2		(int -> goal) -> (int -> prf) -> o.

p1 OutGoal P :- p property1 OutGoal P.
p1a OutGoal P :- p property1a OutGoal P.
p1b OutGoal P :- p property1b OutGoal P.
p2 OutGoal P :- p property2 OutGoal P.

/* ********************************************************************
 * Continue a proof by first reading input from a file, executing it,
 * calling the interactive prover, and then writing the proof to a file.
 ******************************************************************** */

type	continue_proof	string -> goal -> goal -> o.
local	process_script	list (goal -> goal -> o) -> goal -> goal -> o.
local	process_script_aux   list (goal -> goal -> o) -> goal ->
                              list (goal -> goal -> o) -> goal -> o.

continue_proof Name InGoal OutGoal :-
  FName is Name ^ ".hyll",
  get_all_tacs FName Tacs,
  process_script Tacs InGoal MidGoal,
  maptac inter MidGoal OutGoal.

process_script Tacs InGoal OutGoal :-
  process_script_aux Tacs InGoal Rest OutGoal, !, Rest = nil.

process_script_aux nil Goal nil Goal :- !.
process_script_aux Tacs truegoal Tacs truegoal :- !.
process_script_aux InTacs (andgoal InGoal1 InGoal2) OutTacs OutGoal :-
  process_script_aux InTacs InGoal1 MidTacs MidGoal1,
  goalreduce (andgoal MidGoal1 InGoal2) MidGoal2,
  process_script_aux MidTacs MidGoal2 OutTacs OutGoal.
process_script_aux InTacs (allgoal InGoal) OutTacs OutGoal :-
  pi T\ (process_script_aux InTac (InGoal T) OutTacs (MidGoal T)),
  goalreduce (allgoal MidGoal) OutGoal.
process_script_aux (Tac::Rest) InGoal OutTacs OutGoal :-
  Tac InGoal MidGoal1,
  goalreduce MidGoal1 MidGoal2,
  process_script_aux Rest MidGoal2 OutTacs OutGoal.

/* ********************************************************************
 * Main interactive prover
 ******************************************************************** */

type	inter_top	sname -> prf -> goal -> o.
type	inter		goal -> goal -> o.
type	do		o -> goal -> goal -> o.
type	quitg		goal -> goal -> o.
type	backup		goal -> goal -> o.
local	print_form_list	list judg -> int -> o.
local	print_sig_list	list sig_item -> int -> o.
local   nl              o.
local   write           A -> o.
local   process_input   (goal -> goal -> o) -> goal -> goal -> o.

inter_top Name P OutGoal :- sequent Name Seq,
  inter (P # Seq) OutGoal.

inter (P # seq Sig G D J) NewGoal :-
  nl, print "Signature:",
  nl, print_sig_list Sig 1,
  nl, print "Intuitionistic Assumptions: ",
  nl, print_form_list G 1,
  nl, print "Linear Assumptions: ",
  nl, print_form_list D 1,
  nl, print "Conclusion: ",
  nl, write J, nl,
  print "Enter tactic: ",
  readterm std_in Tac,
  process_input Tac (P # seq Sig G D J) NewGoal.

process_input backup _ _ :- !, fail.
process_input quitg NewGoal NewGoal :- !.
process_input (do G) OldGoal NewGoal :- G, inter OldGoal NewGoal.
process_input Tac OldGoal NewGoal :-
  Tac OldGoal MidGoal, maptac inter MidGoal NewGoal.
process_input _ OldGoal NewGoal :- 
  inter OldGoal NewGoal. 

print_form_list nil N.
print_form_list (A::Tail) N :-
  write N, print " ", write A, nl,
  (N1 is (N + 1)),
  print_form_list Tail N1.

print_sig_list nil N.
print_sig_list (A::Tail) N :-
  write N, print " ", write A, nl,
  (N1 is (N + 1)),
  print_sig_list Tail N1.

write A :- term_to_string A Str, print Str.
nl :- print "\n".

/* ********************************************************************
 * Get the list of tactics in file FName
 ******************************************************************** */

type    get_all_tacs    string -> list (goal -> goal -> o) -> o.
local   get_rest_tacs   in_stream -> list (goal -> goal -> o) -> o.
local   get_next_tac    in_stream -> (goal -> goal -> o) -> o.
local   get_rest_tac    in_stream -> string -> (goal -> goal -> o) -> o.
local   get_last_char   string -> string -> o.

get_all_tacs FName Tacs :-
  open_in FName InStr,
  get_rest_tacs InStr Tacs,
  close_in InStr.

get_rest_tacs InStr Tacs :- eof InStr, !, Tacs = nil.
get_rest_tacs InStr (Tac::Rest) :-
  get_next_tac InStr Tac,
  get_rest_tacs InStr Rest.

get_next_tac InStr Tac :-
  input_line InStr Str,
  get_rest_tac InStr Str Tac.

get_rest_tac InStr Str Tac :- Str = "\n", !, Tac = idtac.
get_rest_tac InStr Str Tac :-
  get_last_char Str Chr, Chr = ".", !,  string_to_term Str Tac.
get_rest_tac InStr Str1 Tac :-
  input_line InStr Str2, Str is Str1 ^ Str2, get_rest_tac InStr Str Tac.

get_last_char Str Chr :-
  N is size Str,
  N2 is N - 2,
  Chr is substring Str N2 1.

/* ********************************************************************
 * Translating proofs to strings for use as input to Coq Hyll prover
 ******************************************************************** */

type   write_proof		prf -> judg -> string -> o.
type   proof_to_Coq_string	prf -> judg -> string -> o.
local  auto_str                 string -> o.
local  judg_list_to_string      list judg -> string -> o.
local  judg_to_string           judg -> string -> o.
local  form_to_string           form -> string -> o.
local  form_absentto_string       (A -> form) -> string -> o.
local  world_to_string          int -> string -> o.
local  judg_to_rule_and_world   judg -> (form -> o) -> int -> o.
local  rule_to_string           (form -> o) -> string -> o.

write_proof P J Name :-
  proof_to_Coq_string P J S,
  FName is Name ^ ".v",
  open_out FName OutStr,
  output OutStr S,
  close_out OutStr.

auto_str Str :-
  Str is "; auto.\nunfold munion,madd; simpl; auto.\n" ^
         "apply ms_eq_cons; try (apply eqj_refl);\n" ^
         "  try (apply eq_nat_refl); try (apply eqatm_refl); eauto.\n" ^
         "apply ms_perm___; try (apply eqj_refl);\n" ^
         "  try (apply eq_nat_refl); try (apply eqatm_refl); eauto.\n" ^
         "repeat split; right; auto; intros [hh _]; " ^
         " repeat (inversion_clear 0).\n\n".

proof_to_Coq_string s_init J Str :- !, Str = "apply s_init; auto.\n".
proof_to_Coq_string (s_copy (A @ W) D Prf) J Str :- !,
  proof_to_Coq_string Prf J SPrf,
  form_to_string A SA,
  world_to_string W SW,
  judg_list_to_string D SD,
  Str is "apply s_copy with\n (" ^ SA ^ ")\n (" ^ SW ^ ") \n " ^
         SD ^ "; simpl; auto.\n" ^ SPrf.
proof_to_Coq_string (s_copy J' D Prf) J Str :- !,
  judg_to_rule_and_world J' R W,
  rule_to_string R SR,
  world_to_string W SW,
  judg_list_to_string D SD,
  proof_to_Coq_string Prf J SPrf,
  Str is "apply s_copy with\n (dagger " ^ SR ^ ")\n (" ^ SW ^ ")\n " ^
         SD ^ "; simpl; auto.\n" ^ SPrf.
proof_to_Coq_string (s_ConjMR D1 Prf1 D2 Prf2) (A *o B @ W) Str :- !,
  proof_to_Coq_string Prf1 (A @ W) SPrf1,
  proof_to_Coq_string Prf2 (B @ W) SPrf2,
  judg_list_to_string D1 SD1,
  judg_list_to_string D2 SD2,
  auto_str AutoStr,
  Str is "apply s_ConjMR with\n " ^ SD1 ^ "\n " ^ SD2 ^ AutoStr ^
         SPrf1 ^ SPrf2.
proof_to_Coq_string (s_ConjML A B U D2 D' Prf) J Str :- !,
  proof_to_Coq_string Prf J SPrf,
  form_to_string A SA,
  form_to_string B SB,
  world_to_string U SU,
  judg_list_to_string D2 SD2,
  judg_list_to_string D' SD',
  auto_str AutoStr,
  Str is "apply s_ConjML with\n (" ^ SA ^ ")\n (" ^ SB ^ ")\n (" ^
         SU ^ ") \n " ^ SD2 ^ "\n " ^ SD' ^ AutoStr ^ SPrf.
proof_to_Coq_string (s_ImpR D' Prf) (A ->> B @ W) Str :- !,
  proof_to_Coq_string Prf (B @ W) SPrf,
  form_to_string A SA,
  form_to_string B SB,
  world_to_string W SW,
  judg_list_to_string D' SD',
  auto_str AutoStr,
  Str is "apply s_ImpR with\n" ^ SD' ^ AutoStr ^ SPrf.
proof_to_Coq_string (s_ImpL A B U D1 D2 D3 Prf1 Prf2) J Str :- !,
  proof_to_Coq_string Prf1 (A @ U) SPrf1,
  proof_to_Coq_string Prf2 J SPrf2,
  form_to_string A SA,
  form_to_string B SB,
  world_to_string U SU,
  judg_list_to_string D1 SD1,
  judg_list_to_string D2 SD2,
  judg_list_to_string D3 SD3,
  auto_str AutoStr,
  Str is "apply s_ImpL with\n (" ^ SA ^ ")\n (" ^ SB ^ ")\n (" ^
         SU ^ ") \n " ^
         SD1 ^ "\n " ^ SD2 ^ "\n " ^ SD3 ^ AutoStr ^ SPrf1 ^ SPrf2.
proof_to_Coq_string (s_ConjAR Prf1 Prf2) (A &a B @ W) Str :- !,
  proof_to_Coq_string Prf1 (A @ W) SPrf1,
  proof_to_Coq_string Prf2 (B @ W) SPrf2,
  Str is "apply s_ConjAR; auto.\n" ^ SPrf1 ^ SPrf2.
proof_to_Coq_string (s_ConjAL1 A B U DRest D' Prf) J Str :- !,
  proof_to_Coq_string Prf J SPrf,
  form_to_string A SA,
  form_to_string B SB,
  world_to_string U SU,
  judg_list_to_string DRest SDRest,
  judg_list_to_string D' SD',
  auto_str AutoStr,
  Str is "apply s_ConjAL1 with\n (" ^ SA ^ ")\n (" ^ SB ^ ")\n (" ^
         SU ^ ") \n " ^ SDRest ^ "\n " ^ SD' ^ AutoStr ^ SPrf.
proof_to_Coq_string (s_ConjAL2 A B U DRest D' Prf) J Str :- !,
  proof_to_Coq_string Prf J SPrf,
  form_to_string A SA,
  form_to_string B SB,
  world_to_string U SU,
  judg_list_to_string DRest SDRest,
  judg_list_to_string D' SD',
  auto_str AutoStr,
  Str is "apply s_ConjAL2 with\n (" ^ SA ^ ")\n (" ^ SB ^ ")\n (" ^
         SU ^ ") \n " ^ SDRest ^ "\n " ^ SD' ^ AutoStr ^ SPrf.
proof_to_Coq_string (s_DisjAR1 Prf) (A +o B @ W) Str :- !,
  proof_to_Coq_string Prf (A @ W) SPrf,
  Str is "apply s_DisjAR1; auto.\n" ^ SPrf.
proof_to_Coq_string (s_DisjAR2 Prf) (A +o B @ W) Str :- !,
  proof_to_Coq_string Prf (B @ W) SPrf,
  Str is "apply s_DisjAR2; auto.\n" ^ SPrf.
proof_to_Coq_string (s_DisjAL A B U DRest D1 D2 Prf1 Prf2) J Str :- !,
  proof_to_Coq_string Prf1 J SPrf1,
  proof_to_Coq_string Prf2 J SPrf2,
  form_to_string A SA,
  form_to_string B SB,
  world_to_string U SU,
  judg_list_to_string DRest SDRest,
  judg_list_to_string D1 SD1,
  judg_list_to_string D2 SD2,
  auto_str AutoStr,
  Str is "apply s_DisjAL with\n (" ^ SA ^ ")\n (" ^ SB ^ ")\n (" ^
         SU ^ ") \n " ^
         SDRest ^ "\n " ^ SD1 ^ "\n " ^ SD2 ^ AutoStr ^ SPrf1 ^ SPrf2.
proof_to_Coq_string (s_WallL A U V DRest D' Prf) J Str :-
  proof_to_Coq_string Prf J SPrf,
  form_absentto_string A SA,
  world_to_string U SU,
  world_to_string V SV,
  judg_list_to_string DRest SDRest,
  judg_list_to_string D' SD',
  auto_str AutoStr,
  Str is "apply s_WallL with\n (" ^ SA ^ ")\n (" ^ SU ^ ")\n (" ^
         SV ^ ")\n " ^ SDRest ^ "\n " ^ SD' ^ AutoStr ^ SPrf.
proof_to_Coq_string (s_BangL A U D' Prf) J Str :-
  proof_to_Coq_string Prf J SPrf,
  form_to_string A SA,
  world_to_string U SU,
  judg_list_to_string D' SD',
  auto_str AutoStr,
  Str is "apply s_BangL with\n (" ^ SA ^ ")\n (" ^ SU ^ ")\n " ^ SD' ^
         AutoStr ^ SPrf.
proof_to_Coq_string (s_AtR Prf) ((A at U) @ W) Str :- !,
  proof_to_Coq_string Prf (A @ U) SPrf,
  Str is "apply s_AtR; auto.\n" ^ SPrf.
proof_to_Coq_string (s_AtL A U V DRest D' Prf) J Str :-
  proof_to_Coq_string Prf J SPrf,
  form_to_string A SA,
  world_to_string U SU,
  world_to_string V SV,
  judg_list_to_string DRest SDRest,
  judg_list_to_string D' SD',
  auto_str AutoStr,
  Str is "apply s_AtL with\n (" ^ SA ^ ")\n (" ^ SU ^ ")\n (" ^
         SV ^ ")\n " ^ SDRest ^ "\n " ^ SD' ^ AutoStr ^ SPrf.
proof_to_Coq_string (s_DownR Prf) (down A @ W) Str :- !,
  proof_to_Coq_string Prf (A W @ W) SPrf,
  Str is "apply s_DownR; auto.\n" ^ SPrf.
proof_to_Coq_string (s_DownL A V DRest D' Prf) J Str :-
  proof_to_Coq_string Prf J SPrf,
  form_absentto_string A SA,
  world_to_string V SV,
  judg_list_to_string DRest SDRest,
  judg_list_to_string D' SD',
  auto_str AutoStr,
  Str is "apply s_DownL with\n (" ^ SA ^ ")\n (" ^
         SV ^ ")\n " ^ SDRest ^ "\n " ^ SD' ^ AutoStr ^ SPrf.
proof_to_Coq_string (s_stepR Prf) (step A @ W) Str :- !,
  proof_to_Coq_string Prf (down (w\ (A at (w + 1))) @ W) SPrf,
  Str = "unfold step; auto.\n" ^ SPrf.
proof_to_Coq_string (s_unfold_l Name Prf) J Str :- !,
  proof_to_Coq_string Prf J SPrf,
  rule_to_string R SR,
  Str = "unfold " ^ Name ^ "; auto.\n" ^ SPrf.
proof_to_Coq_string (s_unfold_r Name J' Prf) J Str :- !,
  proof_to_Coq_string Prf J' SPrf,
  rule_to_string R SR,
  Str = "unfold " ^ Name ^ "; auto.\n" ^ SPrf.
proof_to_Coq_string (d_unfold R Prf) J Str :- !,
  proof_to_Coq_string Prf J SPrf,
  rule_to_string R SR,
  Str = "unfold " ^ SR ^ "; auto.\n" ^ SPrf.
proof_to_Coq_string P J Str :- Str = "",
  term_to_string P PStr, print PStr, print "\n",
  term_to_string J JStr, print JStr, print "\n".

judg_to_rule_and_world (fdef R (W1\ dagger W1 @ W)) R W :- !.

rule_to_string rule1 "rule1".
rule_to_string rule2 "rule2".
rule_to_string rule3 "rule3".
rule_to_string rule4 "rule4".
rule_to_string rule5 "rule5".
rule_to_string rule6 "rule6".
rule_to_string s_rule1 "s_rule1".
rule_to_string s_rule2 "s_rule2".
rule_to_string s_rule3 "s_rule3".
rule_to_string s_rule4 "s_rule4".
rule_to_string s_rule5 "s_rule5".
rule_to_string s_rule6 "s_rule6".

judg_list_to_string nil "mnil_".
judg_list_to_string ((fdef R A)::Js) Str :- !,
  judg_to_rule_and_world (fdef R A) R W,
  rule_to_string R SR,
  world_to_string W SW,
  judg_list_to_string Js SJs,
  Str = "(madd (dagger " ^ SR ^ " @ " ^ SW ^ ")\n  " ^ SJs ^ ")".
judg_list_to_string (J::Js) Str :-
  judg_list_to_string Js SJs,
  judg_to_string J SJ,
  Str = "(madd (" ^ SJ ^ ")\n  " ^ SJs ^ ")".

judg_to_string J Str :- term_to_string J Str.
form_to_string A Str :- term_to_string A Str.
form_absentto_string A Str :- term_to_string A Str.
world_to_string W Str :- term_to_string W Str.
